-- Attach it to the common Library that Zorba seems to be aiming for
if not Library then Library = {} end

--[[
Todo
- Column Color?
- Font Color
- Font selection
--]]

-- Declare a basic unit the cell object
local Cell = class()
function Cell:__init(parent)
    -- Instantiate a new background frame
    self.frame = Frames:Create("Frame", "Cell", parent.frame)

    -- Remember the parent
    self.parent = parent

    -- Inherit color from row
    self.bgColor = parent.bgColor
    self.frame:SetBackgroundColor(unpack(self.bgColor))

    -- Inherit default justification from parent
    self.hAlignment = parent.hAlignment
    self.vAlignment = parent.vAlignment

    self.frame:SetVisible(true)
end

function Cell:SetText(text)
    -- Just recycle in case it was previously a different frame type
    if self.valueFrame ~= nil then
        Frames:Recycle(self.valueFrame)
    end

    -- Instantiate a new frame on top of the cell background that contains the actual text
    self.valueFrame = Frames:Create("Text", "CellContents", self.frame)

    -- Set the text
    self.text = text
    self.valueFrame:SetText(self.text)

    -- Set up valueFrame attachments to the cell frame
    self:SetAlignment(self.hAlignment, self.vAlignment)

    self.valueFrame:SetVisible(true)
end

function Cell:SetTexture(source, resource, maxHeight, maxWidth)
    -- Just recycle in case it was previously a different frame type
    if self.valueFrame ~= nil then
        Frames:Recycle(self.valueFrame)
    end

    -- Use these to specify the texture's desired height/width
    self.maxHeight = maxHeight
    self.maxWidth = maxWidth

    -- Instantiate a new frame on top of the cell background that contains the actual texture
    self.valueFrame = Frames:Create("Texture", "CellContents", self.frame)

    -- Set the texture
    self.source = source
    self.resource = resource
    self.valueFrame:SetTexture(self.source, self.resource)

    -- Set up valueFrame attachments to the cell frame
    self:SetAlignment("CENTER", "CENTER")

    self.valueFrame:SetVisible(true)
end

function Cell:GetPreferredHeight()
    if not self.valueFrame then
        return nil
    elseif self.valueFrame.GetTextureHeight then
        -- Must be a texture
        -- We want the text to resize the texture not the other way round
        return self.maxHeight
    end
end

function Cell:SetBackgroundColor(r, g, b, a)
    self.bgColor = {r, g, b, a}
    self.frame:SetBackgroundColor(r, g, b, a)
end

function Cell:SetAlignment(hAlignment, vAlignment)
    local frame = self.valueFrame
    if not frame then
        return
    end

    frame:ClearAll()

    local aWord = vAlignment .. hAlignment
    local parentFrame = self.frame

    frame:SetPoint(aWord, parentFrame, aWord)

    self.hAlignment = hAlignment
    self.vAlignment = vAlignment
end

function Cell:RecycleFrames()
    if self.valueFrame then
        self.valueFrame:SetVisible(false)
        Frames:Recycle(self.valueFrame)
        self.valueFrame = nil
    end

    if self.frame then
        self.frame:SetVisible(false)
        Frames:Recycle(self.frame)
        self.frame = nil
    end
end

function Cell:SetValueFrameHeight(rowHeight)
    if self.valueFrame then
        -- Use the max height if provided else fall back on the row height
        if self.valueFrame.GetTexture then
            if self.maxHeight then
                self.valueFrame:SetHeight(self.maxHeight)
            else
                self.valueFrame:SetHeight(rowHeight)
            end
        end
    end
end

-- Declare a collection of cell objects the row object
local Row = class()
function Row:__init(nColumns, parent)
    -- Instantiate a new row frame
    self.frame = Frames:Create("Frame", "RowFrame", parent.frame)

    -- Remember the parent
    self.parent = parent

    -- Set up row frame attachments
    local previousRow = parent.rows[# parent.rows]
    if not previousRow then
        self.frame:SetPoint("TOPLEFT", parent.frame, "TOPLEFT")
        self.frame:SetPoint("TOPRIGHT", parent.frame, "TOPRIGHT")
    else
        self.frame:SetPoint("TOPLEFT", previousRow.frame, "BOTTOMLEFT")
        self.frame:SetPoint("TOPRIGHT", previousRow.frame, "BOTTOMRIGHT")
    end

    -- ***Set all inherited characteristics before we create the cell frames***

    -- Inherit the default color from the table
    self.bgColor = parent.bgColor
    self.frame:SetBackgroundColor(unpack(self.bgColor))

    -- Inherit default justification from parent
    self.hAlignment = parent.hAlignment
    self.vAlignment = parent.vAlignment

    -- Set up the cells in the row
    local cells = {}
    local cell
    for i = 1, nColumns do
        cell = Cell(self)
        table.insert(cells, cell)

        -- Set up cell frame attachments
        if i == 1 then
            cell.frame:SetPoint("TOPLEFT", self.frame, "TOPLEFT")
            cell.frame:SetPoint("BOTTOMLEFT", self.frame, "BOTTOMLEFT")
        else
            cell.frame:SetPoint("TOPLEFT", cells[i-1].frame, "TOPRIGHT")
            cell.frame:SetPoint("BOTTOMLEFT", cells[i-1].frame, "BOTTOMRIGHT")
        end
    end
    self.cells = cells

    -- Make the row frame visible
    self.frame:SetVisible(true)
end

function Row:GetColumnWidths()
    return self.parent.columnWidths
end

function Row:AddCell()
    local newCell = Cell(self)
    local prevCell = self.cells[# self.cells]
    table.insert(self.cells, newCell)

    newCell.frame:SetPoint("TOPLEFT", prevCell.frame, "TOPRIGHT")
    newCell.frame:SetPoint("BOTTOMLEFT", prevCell.frame, "BOTTOMRIGHT")

    return newCell
end

function Row:GetCell(colIndex)
    return self.cells[colIndex]
end

function Row:GetPreferredHeight()
    if self.height then
        return self.height
    end

    local maxHeight = 0
    for cellIdx, cell in ipairs(self.cells) do
        if cell:GetPreferredHeight() then
            if maxHeight < cell:GetPreferredHeight() then
                maxHeight = cell:GetPreferredHeight()
            end
        end
    end

    if maxHeight == 0 then
        maxHeight = nil
    end

    return maxHeight
end

function Row:SetHeight(height)
    self.height = height
end

function Row:SetCellFrameHeights(rowHeight)
    -- This will automatically resize the cell background frames as well
    self.frame:SetHeight(rowHeight)

    -- Check if there are any texture value frames that we have to resize
    for cellIdx, cell in ipairs(self.cells) do
        cell:SetValueFrameHeight(rowHeight)
    end
end

local function AutoScale(origX, newX, origY)
    return (origY/origX)*newX
end

function Row:SetCellFrameWidths()
    local columnWidths = self:GetColumnWidths()
    for cellIdx, cell in ipairs(self.cells) do
        cell.frame:SetWidth(columnWidths[cellIdx])
    end

    -- Autoscale widths of texture valueframes
    local vf, newWidth
    for cellIdx, cell in ipairs(self.cells) do
        if cell.valueFrame then
            if cell.valueFrame.GetTexture then
                vf = cell.valueFrame
                newWidth = AutoScale(vf:GetTextureHeight(), vf:GetHeight(), vf:GetTextureWidth())
                vf:SetWidth(newWidth)
            end
        end
    end

end

function Row:SetBackgroundColor(r, g, b, a)
    for cellIdx, cell in ipairs(self.cells) do
        cell:SetBackgroundColor(r, g, b, a)
    end
end

function Row:SetAlignment(hAlignment, vAlignment)
    for cellIdx, cell in ipairs(self.cells) do
        cell:SetAlignment(hAlignment, vAlignment)
    end

    -- Remember for future cells
    self.hAlignment = hAlignment
    self.vAlignment = vAlignment
end

function Row:RecycleFrames()
    for cellIdx, cell in ipairs (self.cells) do
        cell:RecycleFrames()
    end

    if self.frame then
        self.frame:SetVisible(false)
        Frames:Recycle(self.frame)
        self.frame = nil
    end
end

function Row:RemoveColumn(colIndex)
    -- Remove cell and value frames
    local deadCellWalking = self.cells[colIndex]
    if not deadCellWalking then
        return
    end

    deadCellWalking:RecycleFrames()

    -- remove reference to the cell object
    table.remove(self.cells, colIndex)

    -- Reattach the cell frame that was previously attached
    if colIndex <= #self.cells then
        local nextCell = self.cells[colIndex]
        local previousCell = self.cells[colIndex - 1]

        if not previousCell then
            nextCell.frame:SetPoint("TOPLEFT", self.frame, "TOPLEFT")
            nextCell.frame:SetPoint("BOTTOMLEFT", self.frame, "BOTTOMLEFT")
        else
            nextCell.frame:SetPoint("TOPLEFT", previousCell.frame, "TOPRIGHT")
            nextCell.frame:SetPoint("BOTTOMLEFT", previousCell.frame, "BOTTOMRIGHT")
        end
    end

end

-- Declare a collection of row objects the table object
Table = class()
function Table:__init(nColumns, parentFrame)
    -- Remember the parent
    self.parent = parentFrame
    self.rows = {}

    -- Instantiate the table frame
    self.frame = Frames:Create("Frame", "Table", parentFrame)

    -- Set attachments for the table frame to the parent
    self.frame:SetAllPoints(parentFrame)

    -- Set the default background color
    self.bgColor = {0, 0, 0, 0}
    self.frame:SetBackgroundColor(unpack(self.bgColor))

    -- Set the default justification
    self.hAlignment = "LEFT"
    self.vAlignment = "CENTER"

    -- Default column width
    self.columns = nColumns     -- this will be used as the default divisor when setting up a new row
    local colWidth = 0
    colWidth = self:GetDefaultColWidth()
    self.defaultWidth = colWidth

    -- Default column percentages
    local colPercent = 0
    colPercent = 1/nColumns

    -- List of column widths and percentages
    local colWidths = {}
    local colPercents = {}
    for i = 1, nColumns do
        table.insert(colWidths, colWidth)
        table.insert(colPercents, colPercent)
    end
    self.columnWidths = colWidths
    self.columnPercents = colPercents

    self.frame:SetVisible(true)
end

function Table:AddRow()
    local newRow = Row(self.columns, self)
    table.insert(self.rows, newRow)

    return newRow
end

function Table:AddColumn()
    local newColumn = {}
    for rowIdx, row in ipairs(self.rows) do
        table.insert(newColumn, row:AddCell())
    end
    self.columns = self.columns + 1

    table.insert(self.columnWidths, self.defaultWidth)

    return newColumn
end

function Table:GetRowCount()
    return # self.rows
end

function Table:GetColumnCount()
    return self.columns
end

function Table:GetRow(rowIndex)
    return self.rows[rowIndex]
end

function Table:GetColumn(colIndex)
    local column = {}
    local cell
    for rowIdx, row in ipairs(self.rows) do
        cell = row:GetCell(colIndex)
        table.insert(column, cell)
    end

    return column
end

function Table:GetCell(rowIndex, colIndex)
    local row = self:GetRow(rowIndex)
    local cell = row:GetCell(colIndex)
    return cell
end

function Table:GetDefaultRowHeight()
    return 20
end

function Table:GetDefaultColWidth()
    return self.frame:GetWidth()/self.columns
end

function Table:SetBackgroundColor(r, g, b, a)
    for rowIdx, row in ipairs(self.rows) do
        row:SetBackgroundColor(r, g, b, a)
    end
end

function Table:SetAlignment(hAlignment, vAlignment)
    for rowIdx, row in ipairs(self.rows) do
        row:SetAlignment(hAlignment, vAlignment)
    end

    -- Remember for future rows
    self.hAlignment = hAlignment
    self.vAlignment = vAlignment

end

function Table:ResizeColumnWidthsToParent()
    local colWidth = self:GetDefaultColWidth()
    self.defaultWidth = colWidth

    local colWidths = {}
    for i = 1, self.columns do
        table.insert(colWidths, colWidth)
    end
    self.columnWidths = colWidths

end

function Table:ResizeColumnWidth(colIndex, colWidth)
    self.columnWidths[colIndex] = colWidth
end

function Table:RemoveRow(rowIndex)
    -- Recycle row, cell and value frames
    local deadRowWalking = self.rows[rowIndex]
    if not deadRowWalking then
        return
    end

    deadRowWalking:RecycleFrames()

    -- Remove the reference to the row object
    table.remove(self.rows, rowIndex)

    -- Reattach frame that was previously attached to the one we just deleted
    if rowIndex <= #self.rows then
        local nextRow = self.rows[rowIndex]
        local previousRow = self.rows[rowIndex - 1]
        if not previousRow then
            nextRow.frame:SetPoint("TOPLEFT", self.frame, "TOPLEFT")
            nextRow.frame:SetPoint("TOPRIGHT", self.frame, "TOPRIGHT")
        else
            nextRow.frame:SetPoint("TOPLEFT", previousRow.frame, "BOTTOMLEFT")
            nextRow.frame:SetPoint("TOPRIGHT", previousRow.frame, "BOTTOMRIGHT")
        end
    end

end

function Table:RemoveColumn(colIndex)
    -- Loop through rows removing the cell corresponding to that column
    for rowIdx, row in ipairs(self.rows) do
        row:RemoveColumn(colIndex)
    end

    -- Update number of columns and column widths
    self.columns = self.columns - 1
    table.remove(self.columnWidths, colIndex)
end

function Table:SetColumnPercentages(colPercents)
    -- Check that enough data was provided
    if #colPercents ~= self.columns then
        print("SetColumnPercentages: Not enough columns specified. There are " .. self.columns .. " columns, " .. #colPercents .. "percentages were provided.")
    end

    -- Calculate the column width in pixels based on the percentage of the full table width
    local sum = 0
    local tableWidth = self.frame:GetWidth()
    local colWidths = {}
    for colIdx, colPercent in ipairs(colPercents) do
        sum = sum + colPercent
        table.insert(colWidths, tableWidth*colPercent)
    end

    -- Check that percentages add up to 1 and store the values
    if sum ~= 1 then
        print("SetColumnPercentages: Column percentages do not add up to 1: ", sum)
    else
        self.columnWidths = colWidths
        self.columnPercents = colPercents
    end

end

function Table:RefreshLayout()
    local rowHeight = 0

    local totalHeight = 0

    for rowIdx, row in ipairs(self.rows) do
        -- Get the height
        rowHeight = row:GetPreferredHeight()
        if not rowHeight then
            rowHeight = self:GetDefaultRowHeight()
        end
        -- Set the height
        row:SetCellFrameHeights(rowHeight)

        -- Add the height to the totalheight
        if row.frame:GetVisible() then
            totalHeight = totalHeight + rowHeight
        end

        -- Set the widths
        row:SetCellFrameWidths()

    end

    -- Store the totalheight
    self.totalHeight = totalHeight

end

function Table:ReportHeights()
    for rowIdx, row in ipairs(self.rows) do
        for cellIdx, cell in ipairs(row.cells) do
            if cell.valueFrame then
                print("(" .. rowIdx .. ", " .. cellIdx .. ") " .. (cell.valueFrame:GetHeight() or "nil"))
            end
        end
    end

end
